@@ -1,14 +1,27 @@
 import json
+import pytest
 import os
-from collections import defaultdict, OrderedDict
+from collections import defaultdict, OrderedDict, namedtuple
+from typing import TYPE_CHECKING
 
 from _pytest.config import create_terminal_writer
 
+if TYPE_CHECKING:
+    from typing import List, Tuple
+    from _pytest.config.argparsing import Parser
+    from _pytest.main import Session
+
+    from _pytest import nodes
+    from _pytest.config import Config
+
 # Ugly hacks for freezegun compatibility: https://github.com/spulec/freezegun/issues/286
 STORE_DURATIONS_SETUP_AND_TEARDOWN_THRESHOLD = 60 * 10  # seconds
 
+TestGroup = namedtuple("TestGroup", "index, num_tests")
+TestSuite = namedtuple("TestSuite", "splits, num_tests")
 
-def pytest_addoption(parser):
+
+def pytest_addoption(parser: "Parser") -> None:
     group = parser.getgroup(
         "Split tests into groups which execution time is about the same. "
         "Run first the whole suite with --store-durations to save information "
@@ -43,51 +56,102 @@ def pytest_addoption(parser):
     )
 
 
-def pytest_collection_modifyitems(session, config, items):
-    splits = config.option.splits
-    group = config.option.group
-    store_durations = config.option.store_durations
-    durations_report_path = config.option.durations_path
+@pytest.mark.tryfirst
+def pytest_cmdline_main(config: "Config") -> None:
+    group = config.getoption("group")
+    splits = config.getoption("splits")
+
+    if splits is None and group is None:
+        return
+
+    if splits and group is None:
+        raise pytest.UsageError("argument `--group` is required")
+
+    if group and splits is None:
+        raise pytest.UsageError("argument `--splits` is required")
+
+    if splits < 1:
+        raise pytest.UsageError("argument `--splits` must be >= 1")
+
+    if group < 1 or group > splits:
+        raise pytest.UsageError(f"argument `--group` must be >= 1 and <= {splits}")
+
+
+class SplitPlugin:
+    def __init__(self):
+        self._suite: TestSuite
+        self._group: TestGroup
+        self._messages: "List[str]" = []
+
+    def pytest_report_collectionfinish(self, config: "Config") -> "List[str]":
+        lines = []
+        if self._messages:
+            lines += self._messages
+
+        if hasattr(self, "_suite"):
+            lines.append(
+                f"Running group {self._group.index}/{self._suite.splits}"
+                f" ({self._group.num_tests}/{self._suite.num_tests}) tests"
+            )
+
+        prefix = "[pytest-split]"
+        lines = [f"{prefix} {m}" for m in lines]
+
+        return lines
+
+    def pytest_collection_modifyitems(
+        self, config: "Config", items: "List[nodes.Item]"
+    ) -> None:
+        splits = config.option.splits
+        group = config.option.group
+        store_durations = config.option.store_durations
+        durations_report_path = config.option.durations_path
 
-    if any((splits, group)):
-        if not all((splits, group)):
-            return
-        if not os.path.isfile(durations_report_path):
-            return
         if store_durations:
-            # Don't split if we are storing durations
-            return
-    total_tests_count = len(items)
-    if splits and group:
+            if any((group, splits)):
+                self._messages.append(
+                    "Not splitting tests because we are storing durations"
+                )
+            return None
+
+        if not group and not splits:
+            # don't split unless explicitly requested
+            return None
+
+        if not os.path.isfile(durations_report_path):
+            self._messages.append(
+                "Not splitting tests because the durations_report is missing"
+            )
+            return None
+
         with open(durations_report_path) as f:
             stored_durations = OrderedDict(json.load(f))
 
         start_idx, end_idx = _calculate_suite_start_and_end_idx(
             splits, group, items, stored_durations
         )
+
+        self._suite = TestSuite(splits, len(items))
+        self._group = TestGroup(group, end_idx - start_idx)
         items[:] = items[start_idx:end_idx]
 
-        terminal_reporter = config.pluginmanager.get_plugin("terminalreporter")
-        terminal_writer = create_terminal_writer(config)
-        message = terminal_writer.markup(
-            " Running group {}/{} ({}/{} tests)\n".format(
-                group, splits, len(items), total_tests_count
-            )
-        )
-        terminal_reporter.write(message)
+
+def pytest_configure(config: "Config") -> None:
+    config.pluginmanager.register(SplitPlugin())
 
 
-def pytest_sessionfinish(session, exitstatus):
+def pytest_sessionfinish(session: "Session") -> None:
     if session.config.option.store_durations:
         report_path = session.config.option.durations_path
         terminal_reporter = session.config.pluginmanager.get_plugin("terminalreporter")
-        durations = defaultdict(float)
+        durations: dict = defaultdict(float)
         for test_reports in terminal_reporter.stats.values():
             for test_report in test_reports:
                 if hasattr(test_report, "duration"):
                     stage = getattr(test_report, "when", "")
                     duration = test_report.duration
-                    # These ifs be removed after this is solved: https://github.com/spulec/freezegun/issues/286
+                    # These ifs be removed after this is solved:
+                    # https://github.com/spulec/freezegun/issues/286
                     if duration < 0:
                         continue
                     if (
@@ -108,9 +172,13 @@ def pytest_sessionfinish(session, exitstatus):
         terminal_reporter.write(message)
 
 
-def _calculate_suite_start_and_end_idx(splits, group, items, stored_durations):
+def _calculate_suite_start_and_end_idx(
+    splits: int, group: int, items: "List[nodes.Item]", stored_durations: OrderedDict
+) -> "Tuple[int, int]":
     item_node_ids = [item.nodeid for item in items]
-    stored_durations = {k: v for k, v in stored_durations.items() if k in item_node_ids}
+    stored_durations = OrderedDict(
+        {k: v for k, v in stored_durations.items() if k in item_node_ids}
+    )
     avg_duration_per_test = sum(stored_durations.values()) / len(stored_durations)
 
     durations = OrderedDict()
